@@ -1,15 +1,29 @@
 # -*- coding: utf-8 -*-
 
 from fsutil.metadata import (
-    __author__, __copyright__, __description__,
-    __email__, __license__, __title__, __version__, )
+    __author__,
+    __copyright__,
+    __description__,
+    __email__,
+    __license__,
+    __title__,
+    __version__,
+)
 
 import datetime as dt
 import errno
 import glob
 import hashlib
+import json
 import os
-import requests
+
+try:
+    import requests
+
+    requests_installed = True
+except ImportError:
+    requests_installed = False
+
 import shutil
 import sys
 import zipfile
@@ -23,7 +37,15 @@ except ImportError:
 
 
 PY2 = bool(sys.version_info.major == 2)
-SIZE_UNITS = ['bytes', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB']
+SIZE_UNITS = ["bytes", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB"]
+
+
+def _require_requests_installed():
+    if not requests_installed:
+        raise ModuleNotFoundError(
+            "'requests' module is not installed, "
+            "it can be installed by running: 'pip install requests'"
+        )
 
 
 def assert_dir(path):
@@ -31,8 +53,7 @@ def assert_dir(path):
     Raise an OSError if the given path doesn't exist or it is not a directory.
     """
     if not is_dir(path):
-        raise OSError(
-            'Invalid directory path: {}'.format(path))
+        raise OSError("Invalid directory path: {}".format(path))
 
 
 def assert_exists(path):
@@ -40,8 +61,7 @@ def assert_exists(path):
     Raise an OSError if the given path doesn't exist.
     """
     if not exists(path):
-        raise OSError(
-            'Invalid item path: {}'.format(path))
+        raise OSError("Invalid item path: {}".format(path))
 
 
 def assert_file(path):
@@ -49,8 +69,7 @@ def assert_file(path):
     Raise an OSError if the given path doesn't exist or it is not a file.
     """
     if not is_file(path):
-        raise OSError(
-            'Invalid file path: {}'.format(path))
+        raise OSError("Invalid file path: {}".format(path))
 
 
 def assert_not_dir(path):
@@ -58,8 +77,7 @@ def assert_not_dir(path):
     Raise an OSError if the given path is an existing directory.
     """
     if is_dir(path):
-        raise OSError(
-            'Invalid path, directory already exists: {}'.format(path))
+        raise OSError("Invalid path, directory already exists: {}".format(path))
 
 
 def assert_not_exists(path):
@@ -67,8 +85,7 @@ def assert_not_exists(path):
     Raise an OSError if the given path already exists.
     """
     if exists(path):
-        raise OSError(
-            'Invalid path, item already exists: {}'.format(path))
+        raise OSError("Invalid path, item already exists: {}".format(path))
 
 
 def assert_not_file(path):
@@ -76,7 +93,7 @@ def assert_not_file(path):
     Raise an OSError if the given path is an existing file.
     """
     if is_file(path):
-        raise OSError('Invalid path, file already exists: {}'.format(path))
+        raise OSError("Invalid path, file already exists: {}".format(path))
 
 
 def _clean_dir_empty_dirs(path):
@@ -112,11 +129,11 @@ def convert_size_bytes_to_string(size):
     size = float(size)
     units = SIZE_UNITS
     factor = 0
-    factor_limit = (len(units) - 1)
+    factor_limit = len(units) - 1
     while (size >= 1024) and (factor <= factor_limit):
         size /= 1024
         factor += 1
-    s_format = '{:.2f} {}' if (factor > 1) else '{:.0f} {}'
+    s_format = "{:.2f} {}" if (factor > 1) else "{:.0f} {}"
     s = s_format.format(size, units[factor])
     return s
 
@@ -126,7 +143,7 @@ def convert_size_string_to_bytes(size):
     Convert the given size string to bytes.
     """
     units = [item.lower() for item in SIZE_UNITS]
-    parts = size.strip().replace('  ', ' ').split(' ')
+    parts = size.strip().replace("  ", " ").split(" ")
     amount = float(parts[0])
     unit = parts[1]
     factor = units.index(unit.lower())
@@ -165,7 +182,7 @@ def copy_dir_content(path, dest, **kwargs):
     # only if python >= 3.8
     if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
         make_dirs(dest)
-        kwargs.setdefault('dirs_exist_ok', True)
+        kwargs.setdefault("dirs_exist_ok", True)
     shutil.copytree(path, dest, **kwargs)
 
 
@@ -193,7 +210,7 @@ def create_dir(path, overwrite=False):
     make_dirs(path)
 
 
-def create_file(path, content='', overwrite=False):
+def create_file(path, content="", overwrite=False):
     """
     Create file with the specified content at the given path.
     If overwrite is not allowed and path exists, an OSError is raised.
@@ -203,7 +220,9 @@ def create_file(path, content='', overwrite=False):
     write_file(path, content)
 
 
-def create_zip_file(path, content_paths, overwrite=True, compression=zipfile.ZIP_DEFLATED):
+def create_zip_file(
+    path, content_paths, overwrite=True, compression=zipfile.ZIP_DEFLATED
+):
     """
     Create zip file at path compressing directories/files listed in content_paths.
     If overwrite is allowed and dest zip already exists, it will be overwritten.
@@ -213,7 +232,7 @@ def create_zip_file(path, content_paths, overwrite=True, compression=zipfile.ZIP
         assert_not_exists(path)
     make_dirs_for_file(path)
 
-    def _write_content_to_zip_file(file, path, basedir=''):
+    def _write_content_to_zip_file(file, path, basedir=""):
         assert_exists(path)
         if is_file(path):
             filename = get_filename(path)
@@ -226,10 +245,11 @@ def create_zip_file(path, content_paths, overwrite=True, compression=zipfile.ZIP
                     basedir = join_path(basedir, item_name)
                 _write_content_to_zip_file(file, item_path, basedir)
 
-    with zipfile.ZipFile(path, 'w', compression) as file:
+    with zipfile.ZipFile(path, "w", compression) as file:
         for content_path in content_paths:
             _write_content_to_zip_file(file, content_path)
 
+
 def delete_dir(path):
     """
     Alias for remove_dir.
@@ -267,18 +287,21 @@ def delete_files(*paths):
     remove_files(*paths)
 
 
-def download_file(url, dirpath, filename=None, chunk_size=8192):
+def download_file(url, dirpath, filename=None, chunk_size=8192, **kwargs):
     """
     Download a file from url to dirpath.
     If filename is provided, the file will be named using filename.
+    It is possible to pass extra request options (eg. for authentication) using **kwargs.
     """
-    # https://stackoverflow.com/a/16696317/2096218
-    filename = filename or get_filename(url) or 'download'
+    _require_requests_installed()
+    # https://stackoverflow.com/a/16696317/2096218
+    filename = filename or get_filename(url) or "download"
     filepath = join_path(dirpath, filename)
     make_dirs_for_file(filepath)
-    with requests.get(url, stream=True) as response:
+    kwargs["stream"] = True
+    with requests.get(url, **kwargs) as response:
         response.raise_for_status()
-        with open(filepath, 'wb') as file:
+        with open(filepath, "wb") as file:
             for chunk in response.iter_content(chunk_size=chunk_size):
                 if chunk:
                     file.write(chunk)
@@ -300,7 +323,7 @@ def extract_zip_file(path, dest, autodelete=False, content_paths=None):
     """
     assert_file(path)
     make_dirs(dest)
-    with zipfile.ZipFile(path, 'r') as file:
+    with zipfile.ZipFile(path, "r") as file:
         file.extractall(dest, members=content_paths)
     if autodelete:
         remove_file(path)
@@ -330,7 +353,7 @@ def get_dir_creation_date(path):
     return creation_date
 
 
-def get_dir_creation_date_formatted(path, format='%Y-%m-%d %H:%M:%S'):
+def get_dir_creation_date_formatted(path, format="%Y-%m-%d %H:%M:%S"):
     """
     Get the directory creation date formatted using the given format.
     """
@@ -348,16 +371,18 @@ def get_dir_last_modified_date(path):
         for dirname in dirnames:
             dirpath = os.path.join(basepath, dirname)
             last_modified_timestamp = max(
-                last_modified_timestamp, os.path.getmtime(dirpath))
+                last_modified_timestamp, os.path.getmtime(dirpath)
+            )
         for filename in filenames:
             filepath = os.path.join(basepath, filename)
             last_modified_timestamp = max(
-                last_modified_timestamp, os.path.getmtime(filepath))
+                last_modified_timestamp, os.path.getmtime(filepath)
+            )
     last_modified_date = dt.datetime.fromtimestamp(last_modified_timestamp)
     return last_modified_date
 
 
-def get_dir_last_modified_date_formatted(path, format='%Y-%m-%d %H:%M:%S'):
+def get_dir_last_modified_date_formatted(path, format="%Y-%m-%d %H:%M:%S"):
     """
     Get the directory last modification date formatted using the given format.
     """
@@ -406,7 +431,7 @@ def get_file_creation_date(path):
     return creation_date
 
 
-def get_file_creation_date_formatted(path, format='%Y-%m-%d %H:%M:%S'):
+def get_file_creation_date_formatted(path, format="%Y-%m-%d %H:%M:%S"):
     """
     Get the file creation date formatted using the given format.
     """
@@ -422,15 +447,15 @@ def get_file_extension(path):
     return extension
 
 
-def get_file_hash(path, func='md5'):
+def get_file_hash(path, func="md5"):
     """
     Get the hash of the file at the gived path using
     the specified algorithm function (md5 by default).
     """
     assert_file(path)
     hash = hashlib.new(func)
-    with open(path, 'rb') as file:
-        for chunk in iter(lambda: file.read(4096), b''):
+    with open(path, "rb") as file:
+        for chunk in iter(lambda: file.read(4096), b""):
             hash.update(chunk)
     hash_hex = hash.hexdigest()
     return hash_hex
@@ -446,7 +471,7 @@ def get_file_last_modified_date(path):
     return last_modified_date
 
 
-def get_file_last_modified_date_formatted(path, format='%Y-%m-%d %H:%M:%S'):
+def get_file_last_modified_date_formatted(path, format="%Y-%m-%d %H:%M:%S"):
     """
     Get the file last modification date formatted using the given format.
     """
@@ -532,9 +557,9 @@ def join_filename(basename, extension):
     """
     Create a filename joining the file basename and the extension.
     """
-    basename = basename.rstrip('.').strip()
-    extension = extension.replace('.', '').strip()
-    filename = '{}.{}'.format(basename, extension)
+    basename = basename.rstrip(".").strip()
+    extension = extension.replace(".", "").strip()
+    filename = "{}.{}".format(basename, extension)
     return filename
 
 
@@ -552,7 +577,7 @@ def join_path(path, *paths):
     to the directory path of the module in which it's used.
     """
     basepath = path
-    if get_file_extension(path) in ['py', 'pyc', 'pyo']:
+    if get_file_extension(path) in ["py", "pyc", "pyo"]:
         basepath = os.path.dirname(os.path.realpath(path))
     paths = [path.lstrip(os.sep) for path in paths]
     return os.path.normpath(os.path.join(basepath, *paths))
@@ -640,14 +665,14 @@ def move_file(path, dest, overwrite=False, **kwargs):
     shutil.move(path, dest, **kwargs)
 
 
-def read_file(path, encoding='utf-8'):
+def read_file(path, encoding="utf-8"):
     """
     Read the content of the file at the given path using the specified encoding.
     """
     assert_file(path)
-    content = ''
-    options = {} if PY2 else {'encoding': encoding}
-    with open(path, 'r', **options) as file:
+    content = ""
+    options = {} if PY2 else {"encoding": encoding}
+    with open(path, "r", **options) as file:
         content = file.read()
     return content
 
@@ -656,13 +681,39 @@ def read_file_from_url(url, **kwargs):
     """
     Read the content of the file at the given url.
     """
+    _require_requests_installed()
     response = requests.get(url, **kwargs)
     response.raise_for_status()
     content = response.text
     return content
 
 
-def read_file_lines(path, strip_white=True, skip_empty=True, encoding='utf-8'):
+def read_file_json(
+    path,
+    cls=None,
+    object_hook=None,
+    parse_float=None,
+    parse_int=None,
+    parse_constant=None,
+    object_pairs_hook=None,
+):
+    """
+    Read and decode a json encoded file at the given path.
+    """
+    content = read_file(path)
+    data = json.loads(
+        content,
+        cls=cls,
+        object_hook=object_hook,
+        parse_float=parse_float,
+        parse_int=parse_int,
+        parse_constant=parse_constant,
+        object_pairs_hook=object_pairs_hook,
+    )
+    return data
+
+
+def read_file_lines(path, strip_white=True, skip_empty=True, encoding="utf-8"):
     """
     Read file content lines according to the given options.
     """
@@ -775,7 +826,7 @@ def _search_paths(path, pattern):
     """
     assert_dir(path)
     pathname = os.path.join(path, pattern)
-    options = {} if PY2 else {'recursive': True}
+    options = {} if PY2 else {"recursive": True}
     paths = glob.glob(pathname, **options)
     return paths
 
@@ -784,16 +835,14 @@ def search_dirs(path, pattern):
     """
     Search for directories at path matching the given pattern.
     """
-    return _filter_paths(
-        path, _search_paths(path, pattern), predicate=is_dir)
+    return _filter_paths(path, _search_paths(path, pattern), predicate=is_dir)
 
 
 def search_files(path, pattern):
     """
     Search for files at path matching the given pattern.
     """
-    return _filter_paths(
-        path, _search_paths(path, pattern), predicate=is_file)
+    return _filter_paths(path, _search_paths(path, pattern), predicate=is_file)
 
 
 def split_filename(path):
@@ -802,8 +851,8 @@ def split_filename(path):
     """
     filename = get_filename(path)
     basename, extension = os.path.splitext(filename)
-    extension = extension.replace('.', '').strip()
-    return (basename, extension, )
+    extension = extension.replace(".", "").strip()
+    return (basename, extension)
 
 
 def split_filepath(path):
@@ -812,7 +861,7 @@ def split_filepath(path):
     """
     dirpath = os.path.dirname(path)
     filename = get_filename(path)
-    return (dirpath, filename, )
+    return (dirpath, filename)
 
 
 def split_path(path):
@@ -825,12 +874,43 @@ def split_path(path):
     return names
 
 
-def write_file(path, content, append=False, encoding='utf-8'):
+def write_file(path, content, append=False, encoding="utf-8"):
     """
     Write file with the specified content at the given path.
     """
     make_dirs_for_file(path)
-    mode = 'a' if append else 'w'
-    options = {} if PY2 else {'encoding': encoding}
+    mode = "a" if append else "w"
+    options = {} if PY2 else {"encoding": encoding}
     with open(path, mode, **options) as file:
         file.write(content)
+
+
+def write_file_json(
+    path,
+    data,
+    skipkeys=False,
+    ensure_ascii=True,
+    check_circular=True,
+    allow_nan=True,
+    cls=None,
+    indent=None,
+    separators=None,
+    default=None,
+    sort_keys=False,
+):
+    """
+    Write a json file at the given path with the specified data encoded in json format.
+    """
+    content = json.dumps(
+        data,
+        skipkeys=skipkeys,
+        ensure_ascii=ensure_ascii,
+        check_circular=check_circular,
+        allow_nan=allow_nan,
+        cls=cls,
+        indent=indent,
+        separators=separators,
+        default=default,
+        sort_keys=sort_keys,
+    )
+    write_file(path, content)
